"""
Copyright 2025 Ian Housman

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# Module to help with LCMS levels remap crosswalk
# Also assists with collapsing levels

# Object with all levels, class number, color, and name
all_lookup = {
    "Land_Cover": {
        "1": [1, "61BB46", "Vegetated"],
        "2": [2, "58646E", "Non-Vegetated"],
        "3": [3, "1B1716", "Non-Processing Area Mask"],
        "1.1": [1, "004E2B", "Tree Vegetated"],
        "1.2": [2, "8DA463", "Non-Tree Vegetated"],
        "2.1": [3, "893F54", "Non-Vegetated"],
        "3.1": [4, "1B1716", "Non-Processing Area Mask"],
        "1.1.1": [1, "004E2B", "Tree"],
        "1.2.1": [2, "F89A1C", "Shrub"],
        "1.2.2": [3, "E5E98A", "Grass/Forb/Herb"],
        "2.1.1": [4, "893F54", "Barren or Impervious"],
        "2.1.2": [5, "E4F5FD", "Snow or Ice"],
        "2.1.3": [6, "00B6F0", "Water"],
        "3.1.1": [7, "1B1716", "Non-Processing Area Mask"],
        "1.1.1.1": [1, "004E2B", "Tree"],
        "1.1.1.2": [2, "009344", "Tall Shrub & Tree Mix (SEAK Only)"],
        "1.1.1.3": [3, "61BB46", "Shrub & Tree Mix"],
        "1.1.1.4": [4, "ACBB67", "Grass/Forb/Herb & Tree Mix"],
        "1.1.1.5": [5, "8B8560", "Barren & Tree Mix"],
        "1.2.1.1": [6, "CAFD4B", "Tall Shrub (SEAK Only)"],
        "1.2.1.2": [7, "F89A1C", "Shrub"],
        "1.2.1.3": [8, "8FA55F", "Grass/Forb/Herb & Shrub Mix"],
        "1.2.1.4": [9, "BEBB8E", "Barren & Shrub Mix"],
        "1.2.2.1": [10, "E5E98A", "Grass/Forb/Herb"],
        "1.2.2.2": [11, "DDB925", "Barren & Grass/Forb/Herb Mix"],
        "2.1.1.1": [12, "893F54", "Barren or Impervious"],
        "2.1.2.1": [13, "E4F5FD", "Snow or Ice"],
        "2.1.3.1": [14, "00B6F0", "Water"],
        "3.1.1.1": [15, "1B1716", "Non-Processing Area Mask"],
    },
    "Change": {
        "1": [1, "3D4551", "Stable"],
        "2": [2, "D54309", "Loss"],
        "3": [3, "1B1716", "Non-Processing Area Mask"],
        "1.1": [1, "3D4551", "Stable"],
        "1.2": [2, "00A398", "Gain"],
        "2.1": [3, "D54309", "Loss"],
        "3.1": [4, "1B1716", "Non-Processing Area Mask"],
        "1.1.1": [1, "3D4551", "Stable"],
        "1.2.1": [4, "00A398", "Gain"],
        "2.1.1": [2, "F39268", "Slow Loss"],
        "2.1.2": [3, "D54309", "Fast Loss"],
        "3.1.1": [5, "1B1716", "Non-Processing Area Mask"],
    },
    "Land_Use": {
        "1": [1, "FF9EAB", "Anthropogenic"],
        "2": [2, "004E2B", "Non-Anthropogenic"],
        "3": [3, "1B1716", "Non-Processing Area Mask"],
        "1.1": [1, "FBFF97", "Agriculture"],
        "1.2": [2, "E6558B", "Developed"],
        "2.1": [3, "004E2B", "Forest"],
        "2.2": [4, "D4D4D3", "Other"],
        "2.3": [5, "A6976A", "Rangeland or Pasture"],
        "3.1": [6, "1B1716", "Non-Processing Area Mask"],
        "1.1.1": [1, "FBFF97", "Agriculture"],
        "1.2.1": [2, "E6558B", "Developed"],
        "2.1.1": [3, "004E2B", "Forest"],
        "2.2.1": [4, "36C5B2", "Non-Forest Wetland"],
        "2.2.2": [5, "D4D4D3", "Other"],
        "2.3.1": [6, "A6976A", "Rangeland or Pasture"],
        "3.1.1": [7, "1B1716", "Non-Processing Area Mask"],
    },
}


# Get a lookup of MapBiomas names, values, and palette to set as properties for a given band
def getLookup(bandName, codes=all_lookup.keys()):
    return {
        f"{bandName}_class_names": [f"{all_lookup[bandName][c][2]}" for c in codes],
        f"{bandName}_class_palette": [all_lookup[bandName][c][1] for c in codes],
        f"{bandName}_class_values": [all_lookup[bandName][c][0] for c in codes],
    }


# Method to collapse to a given level (1-4), and return the remap values as well as corresponding visualization properties
def getLevelNRemap(level, bandName="Land_Cover"):
    names = all_lookup[bandName].keys()
    level_below = [n for n in names if len(n.split(".")) <= level]
    level_above = [n for n in names if len(n.split(".")) > level]
    the_level = [n for n in names if len(n.split(".")) == level]

    out_remap_below = {all_lookup[bandName][code][0]: all_lookup[bandName][code][0] for code in level_below}
    out_remap_above = {all_lookup[bandName][code][0]: all_lookup[bandName][".".join(code.split(".")[:level])][0] for code in level_above}
    out_remap_below.update(out_remap_above)
    out_remap = out_remap_below

    out_lookup = getLookup(bandName, the_level)
    return {"remap_from": list(out_remap.keys()), "remap_to": list(out_remap.values()), "viz_dict": out_lookup}


# Specify levels to show
product_levels = {"Change": [3, 2, 1], "Land_Cover": [4, 3, 2, 1], "Land_Use": [3, 2, 1]}

if __name__ == "__main__":
    print(getLevelNRemap(2, "Change"))



"""
   Copyright 2025 Ian Housman, Maria Olga Borja

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

# An example of using geeViz to explore the MapBiomas latest Collection outputs
# Adapted from various scripts in https://gee-community-catalog.org/projects/mapbiomas/
####################################################################################################
import os, sys

sys.path.append(os.getcwd())

import ee

# ee.Initialize(project="lcms-292214")

# Module imports
import geeViz.geeView as gv

# Module to help with class codes, names, colors and collapsing levels of MapBiomas data
import geeViz.examples.mapBiomasLookup as mbl

ee = gv.ee
Map = gv.Map
Map.clearMap()
Map.port = 8000


####################################################################################################
# Datasets available here: https://gee-community-catalog.org/projects/mapbiomas/

# Specify which years to show
years = list(range(1985, 2023 + 1))


# Specify projection to use for zonal summaries and map querying
# Be sure to leave one of scale or transform as None
crs = "EPSG:4326"
transform = None
scale = 30

# Choose which level to show (1-4)
# Only 1 and 2 work with on-the-fly Sankey charts
remap_level = 4


Map.setQueryCRS(crs)
if transform == None:
    Map.setQueryScale(scale)
else:
    Map.setQueryTransform(transform)


# Bring in land use land cover datasets and mosaic them
paths = [
    "projects/mapbiomas-public/assets/bolivia/collection2/mapbiomas_bolivia_collection2_integration_v1",  # 1985-2023
    "projects/mapbiomas-public/assets/peru/collection2/mapbiomas_peru_collection2_integration_v1",  # 1985-2022
    "projects/mapbiomas-public/assets/colombia/collection1/mapbiomas_colombia_collection1_integration_v1",  # 1985-2022
    "projects/mapbiomas-public/assets/ecuador/collection1/mapbiomas_ecuador_collection1_integration_v1",  # 1985-2022
    "projects/mapbiomas-public/assets/venezuela/collection1/mapbiomas_venezuela_collection1_integration_v1",  # 1985-2022
    "projects/mapbiomas-public/assets/paraguay/collection1/mapbiomas_paraguay_collection1_integration_v1",  # 1985-2022
    "projects/mapbiomas-public/assets/brazil/lulc/collection9/mapbiomas_collection90_integration_v1",  # 1985-2023
    "projects/mapbiomas-raisg/public/collection5/mapbiomas_raisg_panamazonia_collection5_integration_v1",  # 1985-2022
    "projects/MapBiomas_Pampa/public/collection3/mapbiomas_uruguay_collection1_integration_v1",  # 1985-2022
    "projects/mapbiomas-public/assets/chile/collection1/mapbiomas_chile_collection1_integration_v1",  # 2000-2022
    "projects/mapbiomas-public/assets/argentina/collection1/mapbiomas_argentina_collection1_integration_v1",  # 1998-2022
]

stack = ee.ImageCollection([ee.Image(p).byte() for p in paths]).toBands()

# View palettes source here:
# var palettes = require('users/mapbiomas/modules:Palettes.js');


out_band_name = "lulc"


# Bring in the names, values, and palette
remap_info = mbl.getLevelNRemap(remap_level, out_band_name)


# Function to convert a given band into a time-enabled image object
def setupLulc(yr):

    img = stack.select([f".*_{yr}"]).reduce(ee.Reducer.firstNonNull()).remap(remap_info["remap_from"], remap_info["remap_to"]).rename([out_band_name]).set("system:time_start", ee.Date.fromYMD(yr, 6, 1).millis())
    img = img.set(remap_info["viz_dict"])
    return img


# Convert the image stack into an image collection
lulcC = ee.ImageCollection([setupLulc(yr) for yr in years])

# Only allow Sankey charts for levels <=2
sankey = True
if remap_level > 2:
    sankey = False

# Add the collection to the map
Map.addTimeLapse(lulcC, {"canAreaChart": True, "autoViz": True, "years": years, "areaChartParams": {"line": True, "sankey": sankey, "crs": crs, "transform": transform, "scale": scale}}, "MapBiomas LULC")

# Add layers to summarize by
countries = ee.FeatureCollection("FAO/GAUL_SIMPLIFIED_500m/2015/level0")
Map.addSelectLayer(countries, {}, "Global Country Boundaries")

assetTerritories = ee.FeatureCollection("users/joaovsiqueira1/MAPBIOMAS/ti_uc")
Map.addSelectLayer(assetTerritories, {}, "MapBiomas Territories")


## Set up the map
Map.turnOnAutoAreaCharting()
Map.turnOffLayersWhenTimeLapseIsOn = False
# Map.turnOnInspector()
Map.setCenter(-62.8, -3, 4)

Map.view()



import ee, os, json, re

ee.Initialize()
###################################################################################################
"""
 This script recursively migrates Google Earth Engine assets from one repository to a new one, sets permissions, and deletes old assets if needed.
 **Important** for this to work you need the following:
       - The credential you're using must have editor permissions in both the source and destination repositories
                               OR
       - The credential you're using must have editor permissions in the destination repository, and you are willing to
           change the permissions of all the original assets to "Anyone Can Read" (to do this, set changePermissions = True below)
       - The geeViz package installed in your python instance. (https://github.com/gee-community/geeViz)

"""

# Written by:
# Ian Housman ian.housman@usda.gov ian.housman@gmail.com
# Leah Campbell leah.campbell@usda.gov leahs.campbell@gmail.com
# RedCastle Resources Inc.
# MIT License
###################################################################################################
# Repository you are moving from:
sourceRoot = "users/iwhousman/test"
# sourceRoot = 'projects/USFS/LCMS-NFS/R1/FNF'
# sourceRoot = 'projects/lcms-292214/assets/AK-Ancillary-Data'

# Repository you are moving to:
# It is assumed this folder already exists and user account has write access to it
# destinationRoot = 'projects/lcms-292214/assets/migrationTest/FNF'
# destinationRoot = 'users/usfs-ihousman/migrationTest'
# If the credential you're using does not have editor permissions in the source repository
# Must include 'user:' prefix if it is a individual's Email or 'group:' if it is a Google Group
changeSourcePermissions = True
sourceReaders = []
sourceWriters = ["user:ian.housman@usda.gov"]
source_all_users_can_read = False

changeDestinationPermissions = True
destinationReaders = []
destinationWriters = []
destination_all_users_can_read = False


###################################################################################################
# Function to get all folders, imageCollections, images, and tables under a given folder or imageCollection level
def getTree(fromRoot, toRoot, treeList=[]):
    pathPrefix = "projects/earthengine-legacy/assets/"

    # Clean up the given paths
    if fromRoot[-1] != "/":
        fromRoot += "/"
    if toRoot[-1] != "/":
        toRoot += "/"

    # Handle inconsistencies with the earthengine-legacy prefix
    if re.match("^projects/[^/]+/assets/.*$", fromRoot) == None:
        fromRoot = pathPrefix + fromRoot
    if re.match("^projects/[^/]+/assets/.*$", toRoot) == None:
        toRoot = pathPrefix + toRoot
    # print(fromRoot,toRoot)
    # List assets
    try:
        assets = ee.data.listAssets({"parent": fromRoot})["assets"]
    except Exception as e1:
        print(e1)
        # try:
        #     assets = ee.data.listAssets({'parent':fromRoot})['assets']
        # except Exception as e2:
        #     print(e2)

    # Reursively walk down the tree
    nextLevels = []
    for asset in assets:
        fromID = asset["name"]
        fromType = asset["type"]
        # fromID = fromID.replace('projects/earthengine-legacy/assets/','')
        toID = fromID.replace(fromRoot, toRoot)

        if fromType in ["FOLDER", "IMAGE_COLLECTION"]:
            nextLevels.append([fromID, toID])
        treeList.append([fromType, fromID, toID])

    for i1, i2 in nextLevels:
        getTree(i1, i2, treeList)
    return treeList


###################################################################################################
# Function for setting permissions for all files under a specified root level
# Either a list of assets a root to start from can be provided
def batchChangePermissions(
    assetList=None, root=None, readers=[], writers=[], all_users_can_read=False
):
    if assetList == None:
        assetList = [i[1] for i in getTree(root, root)]

    for assetID in assetList:
        print("Changing permissions for: {}".format(assetID))
        try:
            ee.data.setAssetAcl(
                assetID,
                json.dumps(
                    {
                        "writers": writers,
                        "all_users_can_read": all_users_can_read,
                        "readers": readers,
                    }
                ),
            )
        except Exception as e:
            print(e)


###################################################################################################
# Function to copy all folders, imageCollections, images, and tables under a given folder or imageCollection level
# Permissions can also be set here
def copyAssetTree(
    fromRoot,
    toRoot,
    changePermissions=False,
    readers=[],
    writers=[],
    all_users_can_read=False,
):
    treeList = getTree(fromRoot, toRoot)

    # Iterate across all assets and copy and create when appropriate
    for fromType, fromID, toID in treeList:
        if fromType in ["FOLDER", "IMAGE_COLLECTION"]:
            try:
                print("Creating {}: {}".format(fromType, toID))
                ee.data.createAsset({"type": "Image_Collection", "name": toID})
            except Exception as e:
                print(e)
        else:
            try:
                print("Copying {}: {}".format(fromType, toID))
                ee.data.copyAsset(fromID, toID, False)

            except Exception as e:
                print(e)
        print()
        print()

    if changePermissions:
        batchChangePermissions(
            assetList=[i[2] for i in treeList],
            root=None,
            readers=[],
            writers=[],
            all_users_can_read=False,
        )


###################################################################################################
# Function to delete all folders, imageCollections, images, and tables under a given folder or imageCollection level
def deleteAssetTree(root):
    answer = input(
        "Are you sure you want to delete all assets under {}? (y = yes, n = no) ".format(
            root
        )
    )
    print(answer)
    if answer.lower() == "y":
        answer = input(
            "You answered yes. Just double checking. Are you sure you want to delete all assets under {}? (y = yes, n = no) ".format(
                root
            )
        )
        if answer.lower() == "y":
            treeList = getTree(root, root)
            treeList.reverse()
            for fromType, ID1, ID2 in treeList:
                print("Deleting {}".format(ID1))
                try:
                    ee.data.deleteAsset(ID1)
                except Exception as e:
                    print(e)


###################################################################################################
# Step 1: Make sure source has account added as reader (may need to use different credentials for this)
batchChangePermissions(
    None, sourceRoot, sourceReaders, sourceWriters, source_all_users_can_read
)
###################################################################################################
# Step 2: Use this function to copy assets
# copyAssetTree(sourceRoot,destinationRoot,changeDestinationPermissions,destinationReaders,destinationWriters,destination_all_users_can_read)
###################################################################################################
#!!!!!!! DANGER !!!!!!!!!
#!!!!!!! DANGER !!!!!!!!!
#!!!!!!! DANGER !!!!!!!!!
#!!!!!!! DANGER !!!!!!!!!
# Optional Step 3: Once all assets are copied and inspected, you can use this method to delete all files under the root level
# This method is final and there is no way to undo this
# deleteAssetTree(destinationRoot)



"""
Take data from GEE to Pandas and back

geeViz.gee2Pandas facilitates converting GEE objects to tabular formats that work well in more common packages such as Pandas.  

"""

"""
   Copyright 2025 Ian Housman

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

# --------------------------------------------------------------------------
#           Module to use GEE's online compute capabilities to take data from GEE into more universal environments such as Pandas
# --------------------------------------------------------------------------
import geeViz.geeView
import sys, ee, os, shutil, subprocess, datetime, calendar, json, glob, numpy, pandas
import time, logging, pdb
from simpledbf import Dbf5


#########################################################################
# Function to convert a featureCollection to json
# If the output json file already exists, it will read it in
# Returns the json version of the featureCollection
# Currently maxNumberOfFeatures > 5000 will error out (need to handle slicing a feature list)
def featureCollection_to_json(featureCollection, output_json_name, overwrite=False, maxNumberOfFeatures=5000):
    if not os.path.exists(output_json_name) or overwrite:
        print("Converting featureCollection to json:", os.path.basename(output_json_name))
        t = featureCollection.limit(maxNumberOfFeatures).getInfo()
        o = open(output_json_name, "w")
        o.write(json.dumps(t))
        o.close()
    else:
        print(
            "Already converted featureCollection to json:",
            os.path.basename(output_json_name),
        )
        o = open(output_json_name, "r")
        t = json.load(o)
        o.close()
    return t


#########################################################################
# Convert featureCollection to csv
# Can accept geojson or GEE featureCollection as the featureCollection argument
# Will automatically create geojson version if given a gee FeatureCollection
def featureCollection_to_csv(featureCollection, output_csv_name, overwrite=False):
    if not os.path.exists(output_csv_name) or overwrite:
        df = robust_featureCollection_to_df(featureCollection)
        print("Writing:", output_csv_name)
        df.to_csv(output_csv_name, index=False)
    else:
        print(output_csv_name, " already exists")
        # # First find out if it is a file pathname or GEE featureCollection
        # # Read in if it's a json file pathname
        # # If it is a GEE featureCollection, convert it to json
        # if isinstance(featureCollection,str) and os.path.splitext(featureCollection)[1].lower().find('json') >-1:
        #     o = open(featureCollection,'r')
        #     table = json.load(o)
        #     o.close()
        # elif isinstance(featureCollection,ee.featurecollection.FeatureCollection):
        #     output_json_name = os.path.splitext(output_csv_name)[0]+'.json'
        #     table = featureCollection_to_json(featureCollection,output_json_name,overwrite,maxNumberOfFeatures)

        # # Set up the header
        # # Pull in ID and Lng and Lat field names if it's a point
        # if table['features'][0]['geometry']['type'] == 'Point':
        #     bands = ['ID','Lng','Lat']
        # else:
        #     bands = ['ID']

        # bands.extend(list(table['features'][0]['properties'].keys()))
        # header = ','.join(bands)+'\n'
        # out = header

        # # Iterate across each feature and pull in the id, and properties
        # for feature in table['features']:
        #     props = feature['properties']
        #     geo = feature['geometry']
        #     geoType = geo['type']
        #     ID = str(feature['id'])
        #     values = numpy.array(list(props.values()))

        #     if geoType == 'Point':
        #         ptCoords = geo['coordinates']
        #         values = numpy.concatenate([[ID],geo['coordinates'],values])
        #     values = ','.join([str(i) for i in values])+'\n'
        #     values = values.replace(',None',',')
        #     out+=values

        # o = open(output_csv_name,'w')
        # o.write(out)
        # o.close()


#########################################################################
# Function to convert GEE featureCollection to a Pandas Dataframe
# Handles the 5000 limit
# This will run into memory errors if any complex operations precede the tables creation
# In those instances, the featureCollections should first be exported to asset and then brought back in and run through
# this function
def robust_featureCollection_to_df(featureCollection, sep="___"):
    maxFeatures = 5000
    nFeatures = featureCollection.size().getInfo()
    featureCollection = featureCollection.toList(1000000, 0)
    out_df = []
    for start in range(0, nFeatures, maxFeatures):
        end = start + maxFeatures
        if end > nFeatures:
            end = nFeatures
        print(f"Converting features {start+1}-{end}")
        fcT = ee.FeatureCollection(featureCollection.slice(start, end))
        features = fcT.getInfo()["features"]
        df = pandas.json_normalize(features, sep=sep)
        out_df.append(df)

    out_df = pandas.concat(out_df)

    properties = out_df.columns
    properties_out = [prop.replace(f"properties{sep}", "") for prop in properties]
    properties_out = [prop.replace(f"geometry{sep}", "geometry.") for prop in properties_out]
    prop_dict = dict(zip(properties, properties_out))
    out_df = out_df.rename(columns=prop_dict)

    return out_df


#########################################################################
# Function to compute zonal stats in GEE and download results to a local csv
def geeToLocalZonalStats(
    zones,
    raster,
    output_csv,
    reducer=ee.Reducer.first(),
    scale=None,
    crs=None,
    transform=None,
    tileScale=4,
    overwrite=False,
    maxNumberOfFeatures=5000,
):
    table = raster.reduceRegions(zones, reducer, scale, crs, transform, tileScale)

    featureCollection_to_csv(table, output_csv, overwrite)


#########################################################################
# Function to convert a Pandas dataframe to geojson
# Assumes point location geometry
# Function adapted from: https://notebook.community/captainsafia/nteract/applications/desktop/example-notebooks/pandas-to-geojson
def df_to_geojson(
    df,
    properties=None,
    geometry_type_fieldname="geometry.type",
    geometry_coordinates_fieldname="geometry.coordinates",
):
    # create a new python dict to contain our geojson data, using geojson format
    geojson = {"type": "FeatureCollection", "features": []}

    # Pull all non geo props if none are provided
    if properties == [] or properties == None:
        properties = [col for col in df.columns if col not in [geometry_type_fieldname, geometry_coordinates_fieldname]]

    # loop through each row in the dataframe and convert each row to geojson format
    for _, row in df.iterrows():
        if not pandas.isnull(row[geometry_type_fieldname]) and not pandas.isnull(row[geometry_coordinates_fieldname]):
            # create a feature
            feature = {
                "type": "Feature",
                "properties": {},
                "geometry": {
                    "type": row[geometry_type_fieldname],
                    "coordinates": json.loads(row[geometry_coordinates_fieldname]),
                },
            }

            # for each column, get the value and add it as a new feature property
            for prop in properties:
                p = row[prop]
                if pandas.isnull(p):
                    p = "NA"
                feature["properties"][prop] = p

            # add this feature (aka, converted dataframe row) to the list of features inside our dict
            geojson["features"].append(feature)

    return geojson


####################################################################################################
# Function to take the Excel spreadsheet and convert it to a GEE featureCollection
# Supports Excel (mode='Excel'), CSV (mode='csv'), and Pickle (mode='pickle') input table formats
def tableToFeatureCollection(
    table_path,
    properties=None,
    dateCol=None,
    groupByColumns=None,
    mode=None,
    geometry_type_fieldname="geometry.type",
    geometry_coordinates_fieldname="geometry.coordinates",
):
    mode_dict = {
        ".csv": "csv",
        ".xls": "excel",
        ".xlsx": "excel",
        ".pkl": "pickle",
        ".pickle": "pickle",
    }
    if mode == None:
        try:
            mode = mode_dict[os.path.splitext(table_path)[1]]
        except:
            mode = ""
    # Read in the  table as a Pandas dataframe
    if mode.lower() == "excel":
        df = pandas.read_excel(table_path)
    elif mode.lower() == "csv":
        df = pandas.read_csv(table_path)
    elif mode.lower() == "pickle":
        df = pandas.read_pickle(table_path)
    else:
        raise Exception("Table format not recognized. Support formats are: {}".format(",".join(list(mode_dict.keys()))))

    # Convert the time to a user-friendly format of yyyy-mm-DD
    if dateCol == None:
        for c in df.columns[df.dtypes == "datetime64[ns]"]:
            df[c] = df[c].dt.strftime("%Y-%m-%d")
    else:
        # hcb_df[dateCol]= hcb_df[dateCol].dt.strftime('%Y-%m-%d')
        df[dateCol] = pandas.to_datetime(df[dateCol]).astype(str)

    if groupByColumns != None:
        df = df.groupby(groupByColumns).sum(numeric_only=True).reset_index()

    # Convert the dataframe to geojson
    df_json = df_to_geojson(
        df,
        properties,
        geometry_type_fieldname=geometry_type_fieldname,
        geometry_coordinates_fieldname=geometry_coordinates_fieldname,
    )

    # Read in the geojson as a GEE featureCollection
    df_fc = ee.FeatureCollection(df_json)
    return df_fc


#########################################################################
# Function to convert a dbf to json
def dfToJSON(dbf, outJsonFilename):
    dbf = Dbf5(dbf)
    df = dbf.to_dataframe()  # .head()

    columns = df.columns
    rows = df.transpose().to_numpy()
    outJson = {}
    for i, c in enumerate(columns):
        outJson[c] = list(rows[i])

    o = open(outJsonFilename, "w")
    o.write(json.dumps(outJson))
    o.close()
    return outJson


#########################################################################
# Show image array at a location
def setDFTitle(df, title):
    styles = [
        dict(
            selector="caption",
            props=[
                ("text-align", "left"),
                ("font-size", "150%"),
                ("font-weight", "bold"),
            ],
        )
    ]

    df = df.style.set_caption(title).set_table_styles(styles)
    return df


def imageArrayPixelToDataFrame(
    img,
    pt,
    scale=None,
    crs=None,
    transform=None,
    title=None,
    index=None,
    columns=None,
    bandName=None,
    reducer=ee.Reducer.first(),
    arrayImage=None,
):
    # Pull the values
    vals = img.reduceRegion(reducer, pt, scale=scale, crs=crs, crsTransform=transform).getInfo()

    # Determine if it is an array image
    # Only handle the first band or a single specified band if it is
    if arrayImage == None:
        if type(list(vals.values())[0]) == list:
            arrayImage = True
        else:
            arrayImage = False

    if arrayImage:
        # If no band is specified, and its an array image, pull the first band
        if bandName == None:
            bandName = list(vals.keys())[0]

        # Get the values for the band
        vals = vals[bandName]

        # Convert to Pandas dataframe
        df = pandas.DataFrame(vals, columns=columns, index=index)

    else:
        df = pandas.DataFrame(list(vals.values()), columns=["Values"], index=list(vals.keys()))
    # Set a title if provided
    if title != None:
        df = setDFTitle(df, title)
    return df


# Functions to extract any EE image object type to a dataframe
# Handles multi-band images (with array images as well) and image collections
# While intended for point locations using the ee.Reducer.first() reducer, it can handle polygons and multi-polygons with an appropriate reducer (e.g. ee.Reducer.mean() or ee.Reducer.median())
def extractPointImageValues(
    ee_image,
    pt,
    scale=None,
    crs=None,
    transform=None,
    reducer=ee.Reducer.first(),
    includeNonSystemProperties=False,
    includeSystemProperties=True,
):
    # Pull the values
    ee_image = ee.Image(ee_image)
    system_props = ee_image.toDictionary(["system:index", "system:time_start", "system:time_end"])
    props = ee_image.toDictionary()
    vals = ee.Image(ee_image).reduceRegion(reducer, pt, scale=scale, crs=crs, crsTransform=transform)
    if includeNonSystemProperties and includeSystemProperties:
        props = system_props.combine(props)
        return props.combine(vals)
    elif includeNonSystemProperties and not includeSystemProperties:
        return props.combine(vals)
    elif not includeNonSystemProperties and includeSystemProperties:
        return system_props.combine(vals)
    else:
        return vals


def extractPointValuesToDataFrame(
    ee_object,
    pt,
    scale=None,
    crs=None,
    transform=None,
    title=None,
    index=None,
    columns=None,
    bandName=None,
    reducer=ee.Reducer.first(),
    includeNonSystemProperties=False,
    includeSystemProperties=True,
):
    if isinstance(ee_object, ee.imagecollection.ImageCollection):
        vals = (
            ee_object.toList(10000, 0)
            .map(
                lambda ee_image: extractPointImageValues(
                    ee_image,
                    pt,
                    scale=scale,
                    crs=crs,
                    transform=transform,
                    reducer=reducer,
                    includeNonSystemProperties=includeNonSystemProperties,
                    includeSystemProperties=includeSystemProperties,
                )
            )
            .getInfo()
        )

    elif isinstance(ee_object, ee.image.Image):
        vals = extractPointImageValues(
            ee_object,
            pt,
            scale=scale,
            crs=crs,
            transform=transform,
            reducer=reducer,
            includeNonSystemProperties=includeNonSystemProperties,
            includeSystemProperties=includeSystemProperties,
        ).getInfo()

    return pandas.json_normalize(vals)


####################################################################################################
# Scratch space for testing
if __name__ == "__main__":
    output_dir = r"C:\tmp\geeToPandasTest"
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    #     out_json = os.path.join(output_dir,'test_json.json')

    # out_csv = os.path.join(output_dir,'test_json.csv')
    #     if not os.path.exists(output_dir):os.makedirs(output_dir)
    #     fc = ee.FeatureCollection('projects/lcms-292214/assets/CONUS-LCMS/Training-Tables/Training-Tables-AnnualizedFormat_v2022/LCMS-CONUS-2015').limit(5)
    #     composite = ee.ImageCollection('projects/lcms-tcc-shared/assets/Composites/Composite-Collection-yesL7-1984-2020')\
    #                     .filter(ee.Filter.calendarRange(2015,2015,'year')).mosaic()
    #     # fc = ee.FeatureCollection('projects/lcms-292214/assets/CONUS-Ancillary-Data/FS_District_Boundaries')
    #     # print(fc.size().getInfo())

    #     geeToLocalZonalStats(fc,composite,out_csv,reducer=ee.Reducer.first(),scale=30,crs='EPSG:5070',transform=None,tileScale=4,overwrite=False,maxNumberOfFeatures=500)

    #     fc_test = tableToFeatureCollection(out_csv,lat='Lat', lon='Lng',properties=[],dateCol=None,groupByColumns = None)
    #     print(fc_test.first().getInfo())
    # #     # featureCollection_to_json(fc,out_json,overwrite=True,maxNumberOfFeatures=20)
    # #     # featureCollection_to_csv(fc,out_csv,overwrite = True,maxNumberOfFeatures=500)

    # assets = ee.data.listAssets({'parent': 'projects/rcr-gee/assets/lcms-training/lcms-training_module-4_timeSync'})['assets']

    # training_data = ee.FeatureCollection([ee.FeatureCollection(asset['name']) for asset in assets]).flatten()

    # robust_featureCollection_to_df(training_data.limit(10))
    # featureCollection_to_csv(training_data,out_csv,overwrite = True)

    pt = ee.Geometry.Point([-65.8491, 18.2233])
    comps = ee.ImageCollection("projects/rcr-gee/assets/lcms-training/lcms-training_module-2_composites")
    lt = ee.ImageCollection("projects/rcr-gee/assets/lcms-training/lcms-training_module-3_landTrendr")  # .filter(ee.Filter.eq('band','NBR')).first()\
    # .select(['LandTrendr'])

    # eeObjToDataFrame(ee.Image([1,2,3]),pt,crs = 'EPSG:5070', scale=30,transform = None,title = None,index = None,
    #             columns = None,bandName = None, reducer = ee.Reducer.first())
    # extractPointValues(ee.Image.cat([ee.Image([1,2,3]).toArray(),ee.Image(1)]),pt,crs = 'EPSG:5070', scale=30,transform = None,title = None,index = None,columns = None,bandName = None, reducer = ee.Reducer.first())

    # extracted = extractPointValuesToDataFrame(comps,pt,crs = 'EPSG:5070', scale=30,transform = None,title = None,index = None,columns = None,bandName = None, reducer = ee.Reducer.first(),includeNonSystemProperties = False,includeSystemProperties=True)
    # print(extracted)
    # df = imagePixelToDataFrame(lt,pt,crs = 'EPSG:5070', scale=30,transform = None,index = ['year','vertex fit'],columns = None,bandName = None, reducer = ee.Reducer.first())
    # print(df)
    # df = imagePixelToDataFrame(ee.Image([1,2,3]).byte(),pt,crs = 'EPSG:5070', scale=30,transform = None)

    # print(df)



"""
View GEE objects using Folium

geeViz.foliumView is facilitates viewing GEE objects in Folium. Layers can be added to the map using `Map.addLayer` and then viewed using the `Map.view` method. 

"""

"""
   Copyright 2025 Ian Housman

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
# Script to allow GEE objects to be viewed in folium
# Adapted from: https://colab.research.google.com/github/giswqs/qgis-earthengine-examples/blob/master/Folium/ee-api-folium-setup.ipynb
# Intended to work within the geeViz package
######################################################################
# Import modules
import geeViz.geeView as geeView
import os, sys, folium, json, numpy
from folium import plugins
from IPython.display import display, HTML

# Set up GEE and paths
ee = geeView.ee
######################################################################
# Specify location of files to run
folium_html = "foliumView.html"
folium_html_folder = os.path.join(geeView.py_viz_dir, geeView.geeViewFolder)
######################################################################
# Add custom basemaps to folium
basemaps = {
    "Google Maps": folium.TileLayer(
        tiles="https://mt1.google.com/vt/lyrs=m&x={x}&y={y}&z={z}",
        attr="Google",
        name="Google Maps",
        overlay=True,
        control=True,
    ),
    "Google Satellite": folium.TileLayer(
        tiles="https://mt1.google.com/vt/lyrs=s&x={x}&y={y}&z={z}",
        attr="Google",
        name="Google Satellite",
        overlay=True,
        control=True,
    ),
    "Google Terrain": folium.TileLayer(
        tiles="https://mt1.google.com/vt/lyrs=p&x={x}&y={y}&z={z}",
        attr="Google",
        name="Google Terrain",
        overlay=True,
        control=True,
    ),
    "Google Satellite Hybrid": folium.TileLayer(
        tiles="https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}",
        attr="Google",
        name="Google Satellite",
        overlay=True,
        control=True,
    ),
    "Esri Satellite": folium.TileLayer(
        tiles="https://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{z}/{y}/{x}",
        attr="Esri",
        name="Esri Satellite",
        overlay=True,
        control=True,
    ),
}


######################################################################
# Define a method for displaying Earth Engine image tiles on a folium map.
def add_ee_layer(self, ee_object, vis_params, name, visible):
    try:

        if isinstance(ee_object, ee.image.Image):
            map_id_dict = ee.Image(ee_object).getMapId(vis_params)
            folium.raster_layers.TileLayer(
                tiles=map_id_dict["tile_fetcher"].url_format,
                attr="Google Earth Engine | GeeViz",
                name=name,
                overlay=True,
                control=True,
                show=visible,
            ).add_to(self)
        # display ee.ImageCollection()
        elif isinstance(ee_object, ee.imagecollection.ImageCollection):
            ee_object_new = ee_object.mosaic()
            map_id_dict = ee.Image(ee_object_new).getMapId(vis_params)
            folium.raster_layers.TileLayer(
                tiles=map_id_dict["tile_fetcher"].url_format,
                attr="Google Earth Engine",
                name=name,
                overlay=True,
                control=True,
                show=visible,
            ).add_to(self)
        # display ee.Geometry()
        # elif isinstance(ee_object, ee.geometry.Geometry):
        #     folium.GeoJson(
        #     data = ee_object.getInfo(),
        #     name = name,
        #     overlay = True,
        #     control = True,
        #     show = visible
        # ).add_to(self)
        # display ee.FeatureCollection()
        elif isinstance(ee_object, ee.featurecollection.FeatureCollection) or isinstance(ee_object, ee.feature.Feature) or isinstance(ee_object, ee.geometry.Geometry):
            strokeWidth = 2
            strokeColor = "000"
            if "strokeWidth" in vis_params.keys():
                strokeWidth = vis_params["strokeWidth"]
            if "strokeColor" in vis_params.keys():
                strokeColor = vis_params["strokeColor"]
            vis_params["palette"] = strokeColor
            ee_object_new = ee.Image().paint(ee_object, 0, strokeWidth)
            map_id_dict = ee.Image(ee_object_new).getMapId(vis_params)
            folium.raster_layers.TileLayer(
                tiles=map_id_dict["tile_fetcher"].url_format,
                attr="Google Earth Engine",
                name=name,
                overlay=True,
                control=True,
                show=visible,
            ).add_to(self)

    except Exception as e:
        print("Could not display {}".format(name))
        print("Error:", e)


folium.Map.add_ee_layer = add_ee_layer


######################################################################
# Set up map object
# Intended to be similar to geeView mapper object
class foliumMapper:
    def __init__(self, port=8001):
        self.port = port
        self.isNotebook = geeView.is_notebook()
        self.mapArgs = {}
        self.idDictList = []
        self.layerNumber = 1
        self.mapBounds = None

    def clearMap(self):
        self.layerNumber = 1
        self.idDictList = []
        self.mapBounds = None

    def setMapArg(self, key, value):
        self.mapArgs[key] = value

    def turnOnInspector(self):
        print("Folium Viewer currently only supports viewing of GEE map objects")

    def setCenter(self, lon, lat, zoom=None):
        self.setMapArg("location", [lat, lon])
        if zoom != None:
            self.setMapArg("zoom_start", zoom)

    def centerObject(self, ee_object):
        if isinstance(ee_object, ee.featurecollection.FeatureCollection) or isinstance(ee_object, ee.feature.Feature):
            ee_object = ee_object.geometry()
        bounds = ee_object.bounds(500, "EPSG:4326").coordinates().getInfo()[0]
        xs = [i[0] for i in bounds]
        ys = [i[1] for i in bounds]
        self.mapBounds = [
            [numpy.min(ys), numpy.min(xs)],
            [numpy.max(ys), numpy.max(xs)],
        ]

    def addLayer(self, eeObject, viz={}, name=None, visible=True):
        if name == None:
            name = "Layer " + str(self.layerNumber)
            self.layerNumber += 1
        print("Adding layer: " + name)
        # Get the id and populate dictionary
        idDict = {}  # image.getMapId()
        idDict["item"] = eeObject
        idDict["name"] = name
        idDict["visible"] = visible
        idDict["viz"] = viz

        self.idDictList.append(idDict)

    def turnOffAllLayers(self):
        update = {"visible": False}
        self.idDictList = [{**d, **update} for d in self.idDictList]

    def turnOnAllLayers(self):
        update = {"visible": True}
        self.idDictList = [{**d, **update} for d in self.idDictList]

    def view(self, open_browser=True, open_iframe=False, iframe_height=525):
        self.Map = folium.Map(**self.mapArgs)
        # Add custom basemaps
        if self.mapBounds != None:
            self.Map.fit_bounds(self.mapBounds)

        basemaps["Google Maps"].add_to(self.Map)
        basemaps["Google Satellite Hybrid"].add_to(self.Map)

        for idDict in self.idDictList:
            self.Map.add_ee_layer(idDict["item"], idDict["viz"], idDict["name"], idDict["visible"])

        self.Map.add_child(folium.LayerControl(collapsed=False, autoZIndex=False))

        # Add fullscreen button
        plugins.Fullscreen().add_to(self.Map)

        if not self.isNotebook:
            self.Map.save(os.path.join(folium_html_folder, folium_html))
            if not geeView.isPortActive(self.port):
                print("Starting local web server at: http://localhost:{}/{}/".format(self.port, geeView.geeViewFolder))
                geeView.run_local_server(self.port)
                print("Done")
            else:
                print("Local web server at: http://localhost:{}/{}/ already serving.".format(self.port, geeView.geeViewFolder))
            if open_browser:
                geeView.webbrowser.open(
                    "http://localhost:{}/{}/{}".format(self.port, geeView.geeViewFolder, folium_html),
                    new=1,
                )

        else:
            display(self.Map)
